/*
 * Copyright (c) 2008 - 2012, Andy Bierman, All Rights Reserved.
 * Copyright (c) 2012 - 2020, YumaWorks, Inc., All Rights Reserved.
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 *


    Combined SIL module
    module sethook-example_update
    revision 2020-03-20
    namespace http://netconfcentral.org/ns/sethook-example
    Created: 2020-04-12T20:52:14Z
    CLI parameters:
        format c
        indent 4
        module sethook-example
        output sethook-example.c
        sil-edit2
        sil-get2
        unified true

 */

#include <xmlstring.h>

#include "procdefs.h"
#include "agt.h"
#include "agt_action.h"
#include "agt_cb.h"
#include "agt_cfg.h"
#include "agt_rpc.h"
#include "agt_sil_lib.h"
#include "agt_timer.h"
#include "agt_util.h"
#include "agt_val.h"
#include "cfg.h"
#include "dlq.h"
#include "getcb.h"
#include "ncx.h"
#include "ncx_feature.h"
#include "ncxmod.h"
#include "ncxtypes.h"
#include "obj.h"
#include "rpc.h"
#include "status.h"
#include "val.h"
#include "val_child.h"
#include "val_util.h"
#include "xml_val.h"
#include "sethook-example.h"

/* module static variables */
static ncx_module_t *sethook_example_mod;
static obj_template_t *interfaces_obj;
static obj_template_t *trigger_obj;
static val_value_t *interfaces_val;
static val_value_t *trigger_val;

/* put your static variables here */


/********************************************************************
* FUNCTION print_callback_info
*
* Dump current Callback info
*
********************************************************************/
static void
    print_callback_info (val_value_t *useval,
                         agt_cbtyp_t cbtyp,
                         op_editop_t editop,
                         const xmlChar *callback_name)
{
    if (LOGDEBUG) {
        log_debug("\n (%s) Enter %s callback"
            "\n ------PHASE:%s;EDITOP:%s\n",
            useval ? VAL_NAME(useval) : NCX_NT_NONE,
            callback_name,
            agt_cbtype_name(cbtyp),
            op_editop_name(editop));
    }

} /* print_callback_info */


/********************************************************************
* FUNCTION print_children
*
* Dump children values
*
********************************************************************/
static void print_children (val_value_t *val,
                            const char *which)
{
    val_value_t *child;

    log_debug("\n    %s %s P:[%p] :",
              which,
              val ? VAL_NAME(val) : NCX_EL_NULL,
              val);
    if (val) {
        val_dump_value(val, 10, DBG4);

        child = val_get_first_child(val);

        while (child) {

            if (child->btyp && typ_is_simple(child->btyp)) {
                xmlChar *valstr = val_make_sprintf_string(child);
                if (valstr == NULL) {
                    SET_ERROR(ERR_INTERNAL_MEM);
                    return;
                }

                log_debug("\nchild Q of simple childs:   %s='%s'",
                          VAL_NAME(child),
                          valstr);

                m__free(valstr);
            }

            child = val_get_next_child(child);
        }
    }
} /* print_children */


/********************************************************************
* FUNCTION handle_merge
*
* Dump children values during MERGE operation
*
********************************************************************/
static void handle_merge (rpc_msg_t *msg,
                          val_value_t *newval,
                          val_value_t *curval)
{
    print_children(newval, "newval");
    print_children(curval, "curval");

    log_debug("\n    Merge list:");

    /* the edit is not really on this node; need to get
     * each child_undo record to get the real edited nodes */
    agt_cfg_transaction_t *txcb = RPC_MSG_TXCB(msg);
    agt_cfg_undo_rec_t *child_edit =
        agt_cfg_first_child_edit(txcb, newval, curval);

    while (child_edit) {

        log_debug("\nchild_edit = %p", child_edit);
        log_debug("\nchild_edit->newnode = P:[%p]", child_edit->newnode);
        log_debug("\nchild_edit->curnode = P:[%p]", child_edit->curnode);

        op_editop_t child_editop = OP_EDITOP_NONE;
        val_value_t *child_newval = NULL;
        val_value_t *child_curval = NULL;
        xmlChar *newval_str = NULL;
        xmlChar *curval_str = NULL;

        agt_cfg_child_edit_fields(child_edit,
            &child_editop,
            &child_newval,
            &child_curval);

        if (child_newval) {
            newval_str = val_make_sprintf_string(child_newval);
            if (newval_str == NULL) {
                return;
            }
        }
        if (child_curval) {
            curval_str = val_make_sprintf_string(child_curval);
            if (curval_str == NULL) {
                if (newval_str) {
                    m__free(newval_str);
                }
                return;
            }
        }

        log_debug("\n        %s: editop=%s newval=%s curval=%s",
                  child_newval ? VAL_NAME(child_newval) : NCX_EL_NULL,
                  child_editop ? op_editop_name(child_editop) : NCX_EL_NONE,
                  child_newval ? newval_str : NCX_EL_NULL,
                  child_curval ? curval_str : NCX_EL_NULL);

        if (child_newval && val_set_by_default(child_newval)) {
            log_debug("\n        NEWVAL is default");
        }

        if (child_curval && val_set_by_default(child_curval)) {
            log_debug("\n        CURVAL is default");
        }

        m__free(newval_str);
        m__free(curval_str);

        child_edit = agt_cfg_next_child_edit(child_edit);
    }
} /* handle_merge */


/********************************************************************
* FUNCTION handle_other
*
* Dump values during any but not MERGE operation
*
********************************************************************/
static void handle_other (val_value_t *newval,
                          val_value_t *curval)
{
    print_children(newval, "newval");
    print_children(curval, "curval");

} /* handle_other */


/********************************************************************
* FUNCTION y_sethook_example_init_static_vars
*
* initialize module static variables
*
********************************************************************/
static void y_sethook_example_init_static_vars (void)
{
    sethook_example_mod = NULL;
    interfaces_obj = NULL;
    trigger_obj = NULL;
    interfaces_val = NULL;
    trigger_val = NULL;

    /* init your static variables here */

} /* y_sethook_example_init_static_vars */


/********************************************************************
* FUNCTION sethook_example_edit2
*
* Edit database object callback
*
* Path: container /interfaces/interface/state
* Path: list /interfaces/interface
* Path: container /interfaces
* Path: leaf /trigger

* Add object instrumentation in COMMIT phase.
*
* INPUTS:
*     see agt/agt_cb.h for details
*
* RETURNS:
*     error status
********************************************************************/
static status_t
    sethook_example_edit2 (ses_cb_t *scb,
                           rpc_msg_t *msg,
                           agt_cbtyp_t cbtyp,
                           op_editop_t editop,
                           val_value_t *newval,
                           val_value_t *curval)
{
    status_t res = NO_ERR;
    val_value_t *errorval = (curval) ? curval : newval;
    uint32 cur_errs = agt_get_error_count(msg);

    print_callback_info(errorval, cbtyp, editop, (const xmlChar *)"EDIT-2");

    switch (cbtyp) {
    case AGT_CB_VALIDATE:
        /* description-stmt validation here */
        break;
    case AGT_CB_APPLY:
        /* database manipulation done here */
        break;
    case AGT_CB_COMMIT:
        /* device instrumentation done here */
        switch (editop) {
        case OP_EDITOP_LOAD:
            break;
        case OP_EDITOP_MERGE:
            handle_merge(msg, newval, curval);
            break;
        case OP_EDITOP_REPLACE:
        case OP_EDITOP_CREATE:
        case OP_EDITOP_DELETE:
            handle_other(newval, curval);
            break;
        default:
            res = SET_ERROR(ERR_INTERNAL_VAL);
        }

        if (res == NO_ERR) {
            res = agt_check_cache(&trigger_val, newval, curval, editop);
        }

        break;
    case AGT_CB_ROLLBACK:
        /* undo device instrumentation here */
        break;
    default:
        res = SET_ERROR(ERR_INTERNAL_VAL);
    }

    if ((res != NO_ERR) && (cur_errs == agt_get_error_count(msg))) {
        agt_record_error(
            scb,
            &msg->mhdr,
            NCX_LAYER_CONTENT,
            res,
            NULL,
            (errorval) ? NCX_NT_VAL : NCX_NT_NONE,
            errorval,
            (errorval) ? NCX_NT_VAL : NCX_NT_NONE,
            errorval);
    }
    return res;

} /* sethook_example_edit2 */


/********************************************************************
* FUNCTION create_list_entry
*
* Make a list entry based on the key
*
* INPUTS:
*   res = return status
*
* RETURNS:
*    val_value_t of listval entry if no error
*    else NULL
*
*********************************************************************/
static val_value_t *
    create_list_entry (obj_template_t *list_obj,
                       const xmlChar *keyname,
                       const xmlChar *keystr,
                       status_t *res)
{
    /* add all the /container/list nodes */
    val_value_t *list_value = val_new_value();
    if (!list_value) {
        *res = ERR_NCX_INVALID_VALUE;
        return NULL;
    }
    val_init_from_template(list_value, list_obj);

    /* make key leaf entry */
    val_value_t *child =
        agt_make_leaf(list_obj,
                      keyname,
                      keystr,
                      res);
    if (!child) {
        val_free_value(list_value);
        *res = ERR_NCX_INVALID_VALUE;
        return NULL;
    }
    val_add_child(child, list_value);

    /* generate the internal index Q chain */
    *res = val_gen_index_chain(list_obj, list_value);
    if (*res != NO_ERR) {
        log_error("\nError: could not generate index chain (%s)",
            get_error_string(*res));
        val_free_value(list_value);
        return NULL;
    }

    return list_value;

} /* create_list_entry */


/********************************************************************
* FUNCTION sethook_callback
*
* Callback function for server object handler
* Used to provide a callback for a specific named object
*
* Set Hook:
*   trigger: create /interfaces
*       create container and populate extra nodes
*
*   add_edit effect: populate extra nodes within the container
*        when /interfaces is created
*
*********************************************************************/
static status_t
    sethook_callback (ses_cb_t *scb,
                      rpc_msg_t *msg,
                      agt_cfg_transaction_t *txcb,
                      op_editop_t editop,
                      val_value_t  *newval,
                      val_value_t  *curval)
{
    log_debug("\nEnter SET Hook callback");

    status_t res = NO_ERR;
    val_value_t *errorval = (curval) ? curval : newval;

    /* defpath specified target root */
    const xmlChar *defpath =
        (const xmlChar *)"/sethook-ex:interfaces";

    switch (editop) {
    case OP_EDITOP_LOAD:
        break;
    case OP_EDITOP_MERGE:
    case OP_EDITOP_REPLACE:
        break;
    case OP_EDITOP_CREATE:
        /* populate a new container with several leafs */
        if (newval) {

            /* use newval value to write a new edits */
            val_value_t *editval =
                val_clone_config_data(newval, &res);
            if (editval == NULL) {
                return ERR_INTERNAL_MEM;
            }

            /* add any other children to target container, can add any amount
             * of children here.
             */
            obj_template_t *chobj =
                obj_find_child(VAL_OBJ(editval),
                               y_sethook_example_M_sethook_example,
                               (const xmlChar *)"status");
            if (chobj) {
                val_value_t *child =
                    val_make_simval_obj(chobj,
                                        (const xmlChar *)"extra-leaf",
                                        &res);
                if (child) {
                    res = val_child_add(child, editval);
                }
            } else {
                res = ERR_NCX_NOT_FOUND;
            }

            if (res == NO_ERR) {
                /* add a new edit on defpath and populate new entries */
                res = agt_val_add_edit(scb,
                                       msg,
                                       txcb,
                                       defpath,
                                       editval,
                                       OP_EDITOP_MERGE);
            }

            /* clean up the editval */
            val_free_value(editval);
        }
        break;
    case OP_EDITOP_DELETE:

        break;
    default:
        res = SET_ERROR(ERR_INTERNAL_VAL);
    }

    if (res != NO_ERR) {
        agt_record_error(scb,
                         &msg->mhdr,
                         NCX_LAYER_CONTENT,
                         res,
                         NULL,
                         (errorval) ? NCX_NT_VAL : NCX_NT_NONE,
                         errorval,
                         (errorval) ? NCX_NT_VAL : NCX_NT_NONE,
                         errorval);
    }

    return res;

}  /* sethook_callback */


/********************************************************************
* FUNCTION sethook_callback2
*
* Callback function for server object handler
* Used to provide a callback for a specific named object
*
* Set-Hook: (
*   format: NODE; SUBTREE is NOT supported
*   trigger: CREATE
*      path: /subsystems/subsystem[name='virtual']
*
*   add_edit effect: populate extra container
*        when /subsystems/subsystem is created
*
* INPUTS:
*   scb == session control block making the request
*   msg == incoming rpc_msg_t in progress
*   txcb == transaction control block in progress
*   editop == edit operation enumeration for the node being edited
*   newval == container object holding the proposed changes to
*           apply to the current config, depending on
*           the editop value. Will not be NULL.
*   curval == current container values from the <running>
*           or <candidate> configuration, if any. Could be NULL
*           for create and other operations.
*
* RETURNS:
*    status
*********************************************************************/
static status_t
    sethook_callback2 (ses_cb_t *scb,
                          rpc_msg_t *msg,
                          agt_cfg_transaction_t *txcb,
                          op_editop_t editop,
                          val_value_t  *newval,
                          val_value_t  *curval)
{
    status_t res = NO_ERR;
    val_value_t *errorval = (curval) ? curval : newval;

    handle_other(newval, curval);

    val_value_t *editval = NULL;
    obj_template_t *cont_obj = NULL;
    obj_template_t *list_obj = NULL;

    if (newval) {
        cont_obj =
            obj_find_child(VAL_OBJ(newval),
                           y_sethook_example_M_sethook_example,
                           (const xmlChar *)"worker-threads");

        if (cont_obj) {
            list_obj =
                obj_find_child(cont_obj,
                               y_sethook_example_M_sethook_example,
                               (const xmlChar *)"worker-thread");
        }
    }

    /* defpath specified target root */
    const xmlChar *defpath =
        (const xmlChar *)"/subsystems/subsystem[name='virtual']";

    switch (editop) {
    case OP_EDITOP_LOAD:
        break;
    case OP_EDITOP_MERGE:
    case OP_EDITOP_REPLACE:
        break;
    case OP_EDITOP_CREATE:
        if (newval && cont_obj && list_obj)  {

            log_info("\n\n Start adding Edit value for worker-thread");

            /* clone value in order to write a new edits */
            editval = val_clone_config_data(newval, &res);
            if (!editval) {
                return ERR_NCX_INVALID_VALUE;
            }

            /* now can add more children to the newval */

            val_value_t *cont_val = val_new_value();
            if (!cont_val) {
                return ERR_INTERNAL_MEM;
            }
            val_init_from_template(cont_val, cont_obj);

            /* Create list entries */
            val_value_t *list_value =
                create_list_entry(list_obj,
                                  (const xmlChar *)"thread-name",
                                  (const xmlChar *)"worker-0-1",
                                  &res);
            if (!list_value) {
                log_error("\ncreate_list_entry failed, res %d", res);
                val_free_value(cont_val);
                val_free_value(editval);
                return res;
            }

            res = val_child_add(list_value, cont_val);
            if (res == NO_ERR) {
                res = val_child_add(cont_val, editval);
                if (res == ERR_NCX_DUP_ENTRY) {
                    /* In case NP container already exist (was created by the server
                     * as a default NP-container)
                     * clean up the value before adding a new one
                     */
                    val_value_t *chval =
                        val_find_child_fast(editval,
                                            obj_get_nsid(VAL_OBJ(cont_val)),
                                            obj_get_name(VAL_OBJ(cont_val)));
                    if (chval && val_is_default(chval)) {
                        val_remove_child(chval);
                        val_free_value(chval);
                        chval = NULL;

                        res = val_child_add(cont_val, editval);
                    }
                }
            }

            log_info("\n\n Edit value - added container to value");
            val_dump_value(editval, 3, DBG4);

            /* add a new edit on defpath and populate new entry */
            if (res == NO_ERR) {
                res = agt_val_add_edit(scb,
                                       msg,
                                       txcb,
                                       defpath,
                                       editval,
                                       OP_EDITOP_CREATE);
            }

            /* clean up the editval */
            val_free_value(editval);
        }
        break;
    case OP_EDITOP_DELETE:
        break;
    default:
        res = SET_ERROR(ERR_INTERNAL_VAL);
    }

    if (res != NO_ERR) {
        agt_record_error(scb,
                         &msg->mhdr,
                         NCX_LAYER_CONTENT,
                         res,
                         NULL,
                         (errorval) ? NCX_NT_VAL : NCX_NT_NONE,
                         errorval,
                         (errorval) ? NCX_NT_VAL : NCX_NT_NONE,
                         errorval);
    }
    return res;

}  /* sethook_callback2 */


/********************************************************************
* FUNCTION y_sethook_example_update_init
*
* initialize the sethook-example server instrumentation library
*
* INPUTS:
*    modname == requested module name
*    revision == requested version (NULL for any)
*
* RETURNS:
*     error status
********************************************************************/
status_t
    y_sethook_example_update_init (const xmlChar *modname,
                                   const xmlChar *revision)
{
    status_t res = NO_ERR;

    y_sethook_example_init_static_vars();

    /* change if custom handling done */
    if (xml_strcmp(modname, y_sethook_example_M_sethook_example)) {
        return ERR_NCX_UNKNOWN_MODULE;
    }

    if (revision && xml_strcmp(revision, y_sethook_example_R_sethook_example)) {
        return ERR_NCX_WRONG_VERSION;
    }

    res = ncxmod_load_module(y_sethook_example_M_sethook_example,
                             y_sethook_example_R_sethook_example,
                             agt_get_savedevQ(),
                             &sethook_example_mod);
    if (res != NO_ERR) {
        return res;
    }

    interfaces_obj =
        ncx_find_object(sethook_example_mod,
                        y_sethook_example_N_interfaces);
    if (interfaces_obj == NULL) {
        return ERR_NCX_DEF_NOT_FOUND;
    }

    trigger_obj =
        ncx_find_object(sethook_example_mod,
                        y_sethook_example_N_trigger);
    if (trigger_obj == NULL) {
        return ERR_NCX_DEF_NOT_FOUND;
    }

    res =
        agt_cb_register_edit2_callback(y_sethook_example_M_sethook_example,
                                       (const xmlChar *)"/sethook-ex:interfaces",
                                       y_sethook_example_R_sethook_example,
                                       sethook_example_edit2);
    if (res != NO_ERR) {
        return res;
    }

    res =
        agt_cb_register_edit2_callback(y_sethook_example_M_sethook_example,
                (const xmlChar *)"/sethook-ex:interfaces/sethook-ex:interface",
                                       y_sethook_example_R_sethook_example,
                                       sethook_example_edit2);
    if (res != NO_ERR) {
        return res;
    }

    res =
        agt_cb_register_edit2_callback(y_sethook_example_M_sethook_example,
            (const xmlChar *)"/sethook-ex:interfaces/sethook-ex:interface/sethook-ex:state",
                                       y_sethook_example_R_sethook_example,
                                       sethook_example_edit2);
    if (res != NO_ERR) {
        return res;
    }

    res =
        agt_cb_register_callback(y_sethook_example_M_sethook_example,
                                 (const xmlChar *)"/sethook-ex:trigger",
                                 y_sethook_example_R_sethook_example,
                                 sethook_example_edit2);
    if (res != NO_ERR) {
        return res;
    }

    res =
        agt_cb_register_edit2_callback(y_sethook_example_M_sethook_example,
                                       (const xmlChar *)"/sethook-ex:subsystems",
                                       y_sethook_example_R_sethook_example,
                                       sethook_example_edit2);
    if (res != NO_ERR) {
        return res;
    }

    res =
        agt_cb_register_edit2_callback(y_sethook_example_M_sethook_example,
                    (const xmlChar *)"/sethook-ex:subsystems/sethook-ex:subsystem",
                                       y_sethook_example_R_sethook_example,
                                       sethook_example_edit2);
    if (res != NO_ERR) {
        return res;
    }

    res =
        agt_cb_register_edit2_callback(y_sethook_example_M_sethook_example,
            (const xmlChar *)"/sethook-ex:subsystems/sethook-ex:subsystem/sethook-ex:cpu-resource",
                                       y_sethook_example_R_sethook_example,
                                       sethook_example_edit2);
    if (res != NO_ERR) {
        return res;
    }

    res =
        agt_cb_register_edit2_callback(y_sethook_example_M_sethook_example,
        (const xmlChar *)"/sethook-ex:subsystems/sethook-ex:subsystem/sethook-ex:cpu-resource/sethook-ex:socket",
                                       y_sethook_example_R_sethook_example,
                                       sethook_example_edit2);
    if (res != NO_ERR) {
        return res;
    }

    res =
        agt_cb_register_edit2_callback(y_sethook_example_M_sethook_example,
                (const xmlChar *)"/sethook-ex:subsystems/sethook-ex:subsystem/sethook-ex:cpu-resource/sethook-ex:socket/sethook-ex:cpu",
                                       y_sethook_example_R_sethook_example,
                                       sethook_example_edit2);
    if (res != NO_ERR) {
        return res;
    }

    res =
        agt_cb_register_edit2_callback(y_sethook_example_M_sethook_example,
            (const xmlChar *)"/sethook-ex:subsystems/sethook-ex:subsystem/sethook-ex:worker-threads",
                                       y_sethook_example_R_sethook_example,
                                       sethook_example_edit2);
    if (res != NO_ERR) {
        return res;
    }

    res =
        agt_cb_register_edit2_callback(y_sethook_example_M_sethook_example,
                (const xmlChar *)"/sethook-ex:subsystems/sethook-ex:subsystem/sethook-ex:worker-threads/sethook-ex:worker-thread",
                                       y_sethook_example_R_sethook_example,
                                       sethook_example_edit2);
    if (res != NO_ERR) {
        return res;
    }


    /* put your module initialization code here */


    /* Set Hook callback registration */
    res = agt_cb_hook_register((const xmlChar *)"/sethook-ex:interfaces",
                               AGT_HOOKFMT_NODE,
                               AGT_HOOK_TYPE_SETHOOK,
                               sethook_callback);
    if (res != NO_ERR) {
        return res;
    }


    /* Set Hook to update edits in same transaction */
    res = agt_cb_hook_register((const xmlChar *)"/sethook-ex:subsystems/sethook-ex:subsystem",
                               AGT_HOOKFMT_NODE,
                               AGT_HOOK_TYPE_SETHOOK,
                               sethook_callback2);
    if (res != NO_ERR) {
        return res;
    }

    res = agt_cb_hook_register((const xmlChar *)"/sethook-ex:subsystems/sethook-ex:subsystem/sethook-ex:cpu-resource/sethook-ex:socket/sethook-ex:cpu",
                               AGT_HOOKFMT_NODE,
                               AGT_HOOK_TYPE_SETHOOK,
                               sethook_callback2);
    if (res != NO_ERR) {
        return res;
    }

    return res;

} /* y_sethook_example_update_init */


/********************************************************************
* FUNCTION y_sethook_example_update_init2
*
* SIL init phase 2: non-config data structures
* Called after running config is loaded
*
* RETURNS:
*     error status
********************************************************************/
status_t y_sethook_example_update_init2 (void)
{
    status_t res = NO_ERR;

    interfaces_val =
        agt_init_cache(y_sethook_example_M_sethook_example,
                       y_sethook_example_N_interfaces,
                       &res);
    if (res != NO_ERR) {
        return res;
    }

    trigger_val =
        agt_init_cache(y_sethook_example_M_sethook_example,
                       y_sethook_example_N_trigger,
                       &res);
    if (res != NO_ERR) {
        return res;
    }

    /* put your init2 code here */

    return res;

} /* y_sethook_example_update_init2 */


/********************************************************************
* FUNCTION y_sethook_example_update_cleanup
*    cleanup the server instrumentation library
*
********************************************************************/
void y_sethook_example_update_cleanup (void)
{

    agt_cb_unregister_callbacks(
        y_sethook_example_M_sethook_example,
        (const xmlChar *)"/sethook-ex:interfaces");

    agt_cb_unregister_callbacks(
        y_sethook_example_M_sethook_example,
        (const xmlChar *)"/sethook-ex:interfaces/sethook-ex:interface");

    agt_cb_unregister_callbacks(
        y_sethook_example_M_sethook_example,
        (const xmlChar *)"/sethook-ex:interfaces/sethook-ex:interface/sethook-ex:state");

    agt_cb_unregister_callbacks(
        y_sethook_example_M_sethook_example,
        (const xmlChar *)"/sethook-ex:trigger");

    agt_cb_unregister_callbacks(
        y_sethook_example_M_sethook_example,
        (const xmlChar *)"/sethook-ex:subsystems");

    agt_cb_unregister_callbacks(
        y_sethook_example_M_sethook_example,
        (const xmlChar *)"/sethook-ex:subsystems/sethook-ex:subsystem");

    agt_cb_unregister_callbacks(
        y_sethook_example_M_sethook_example,
        (const xmlChar *)"/sethook-ex:subsystems/sethook-ex:subsystem/sethook-ex:cpu-resource");

    agt_cb_unregister_callbacks(
        y_sethook_example_M_sethook_example,
        (const xmlChar *)"/sethook-ex:subsystems/sethook-ex:subsystem/sethook-ex:cpu-resource/sethook-ex:socket");

    agt_cb_unregister_callbacks(
        y_sethook_example_M_sethook_example,
        (const xmlChar *)"/sethook-ex:subsystems/sethook-ex:subsystem/sethook-ex:cpu-resource/sethook-ex:socket/sethook-ex:cpu");

    agt_cb_unregister_callbacks(
        y_sethook_example_M_sethook_example,
        (const xmlChar *)"/sethook-ex:subsystems/sethook-ex:subsystem/sethook-ex:worker-threads");

    agt_cb_unregister_callbacks(
        y_sethook_example_M_sethook_example,
        (const xmlChar *)"/sethook-ex:subsystems/sethook-ex:subsystem/sethook-ex:worker-threads/sethook-ex:worker-thread");



    /* put your cleanup code here */


    /* Set Hook callback clean up */
    agt_cb_unregister_callbacks(y_sethook_example_M_sethook_example,
                                (const xmlChar *)"/sethook-ex:interfaces");

    agt_cb_unregister_callbacks(y_sethook_example_M_sethook_example,
                                (const xmlChar *)"/sethook-ex:subsystems/sethook-ex:subsystem");

    agt_cb_unregister_callbacks(y_sethook_example_M_sethook_example,
                                (const xmlChar *)"/sethook-ex:subsystems/sethook-ex:subsystem/sethook-ex:cpu-resource/sethook-ex:socket/sethook-ex:cpu");

} /* y_sethook_example_update_cleanup */

/* END SIL sethook_example.c */
